# ruff: noqa: TID252
def ensure_airflow_installed() -> None:
    """Ensures that Airflow is installed."""
    try:
        import airflow  # noqa
    except ImportError:
        raise Exception(
            "Airflow is not installed. Please install Apache Airflow >= 2.0.0 before using this functionality."
            "Airflow has very specific installation instructions, please refer to the official installation guide: "
            "https://airflow.apache.org/docs/apache-airflow/stable/installation/installing-from-pypi.html"
        )


ensure_airflow_installed()
from .base_asset_operator import BaseDagsterAssetsOperator as BaseDagsterAssetsOperator
from .dag_proxy_operator import (
    BaseProxyDAGToDagsterOperator as BaseProxyDAGToDagsterOperator,
    DefaultProxyDAGToDagsterOperator as DefaultProxyDAGToDagsterOperator,
)
from .proxied_state import (
    AirflowProxiedState as AirflowProxiedState,
    DagProxiedState as DagProxiedState,
    TaskProxiedState as TaskProxiedState,
    load_proxied_state_from_yaml as load_proxied_state_from_yaml,
)
from .proxying_fn import proxying_to_dagster as proxying_to_dagster
from .task_proxy_operator import (
    BaseProxyTaskToDagsterOperator as BaseProxyTaskToDagsterOperator,
    DefaultProxyTaskToDagsterOperator as DefaultProxyTaskToDagsterOperator,
)



from collections.abc import Mapping, Sequence

from dagster import JobDefinition
from dagster._core.definitions.assets.definition.asset_spec import AssetSpec
from dagster._core.definitions.decorators.job_decorator import job

from dagster_airlift.core.dag_asset import peered_dag_asset_metadata
from dagster_airlift.core.serialization.serialized_data import (
    SerializedAirflowDefinitionsData,
    SerializedDagData,
)
from dagster_airlift.core.utils import airflow_job_tags, convert_to_valid_dagster_name


def construct_dag_jobs(
    serialized_data: SerializedAirflowDefinitionsData,
    mapped_specs: Mapping[str, Sequence[AssetSpec]],
) -> Sequence[JobDefinition]:
    """Constructs a job for each DAG in the serialized data. The job will be used to power runs."""
    return [
        JobDefinition.for_external_job(
            asset_keys=[spec.key for spec in mapped_specs[dag_id]],
            name=job_name(dag_id),
            metadata=peered_dag_asset_metadata(dag_data.dag_info, dag_data.source_code),
            tags=airflow_job_tags(dag_id),
        )
        for dag_id, dag_data in serialized_data.dag_datas.items()
    ]


def job_name(dag_id: str) -> str:
    """Constructs a job name from the DAG ID. The job name is used to power runs."""
    return convert_to_valid_dagster_name(dag_id)


def dag_non_asset_job(dag_data: SerializedDagData) -> JobDefinition:
    @job(
        name=convert_to_valid_dagster_name(dag_data.dag_id),
        tags=airflow_job_tags(dag_data.dag_id),
    )
    def dummy_job():
        pass

    return dummy_job



import click
from structlog import get_logger

logger = get_logger("dagster-airlift")


@click.group()
def cli():
    """Dagster Airlift CLI. Commands for interacting with the dagster-airlift package."""
    pass


@cli.group()
def proxy() -> None:
    """Commands for working with the Dagster-Airlift proxied state. Requires the `dagster-airlift[in-airflow]` package."""
    try:
        import dagster_airlift.in_airflow  # noqa
    except ImportError:
        raise Exception(
            "dagster-airlift[in-airflow] must be installed in the environment to use any `dagster-airlift proxy` commands."
        )


@proxy.command()
def scaffold():
    """Scaffolds a proxied state folder for the current Airflow installation. Goes in the Airflow Dags folder as <AIRFLOW_DAGS_FOLDER>/proxied_state.

    For each Dag, the proxied state will be filled out as False for all tasks, meaning execution behavior will not be proxied until the user changes the proxied state manually.
    """
    from dagster_airlift.in_airflow.scaffolding import scaffold_proxied_state

    scaffold_proxied_state(logger)


if __name__ == "__main__":
    cli()



import subprocess
from collections.abc import Generator
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Union

from dagster import Component, Resolvable
from dagster._core.definitions.assets.definition.asset_spec import AssetSpec
from dagster._core.definitions.decorators.asset_decorator import asset
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.definitions.job_definition import JobDefinition
from dagster._core.definitions.unresolved_asset_job_definition import UnresolvedAssetJobDefinition
from dagster._core.test_utils import environ
from dagster.components.core.context import ComponentLoadContext
from dagster.components.resolved.core_models import ResolvedAssetKey


def airlift_root() -> Path:
    return Path(__file__).parent.parent.parent


def remove_airflow_home_remnants(airflow_home: Path) -> None:
    logs_path = airflow_home / "logs"
    cfg_path = airflow_home / "airflow.cfg"
    db_path = airflow_home / "airflow.db"
    proxied_state_path = airflow_home / "airflow_dags" / "proxied_state"

    subprocess.check_output(
        ["rm", "-rf", str(logs_path), str(cfg_path), str(db_path), str(proxied_state_path)]
    )


def airflow_cfg_script_path() -> Path:
    return airlift_root() / "scripts" / "airflow_setup.sh"


def chmod_script(script_path: Path) -> None:
    subprocess.check_output(["chmod", "+x", str(script_path)])


@contextmanager
def configured_airflow_home(airflow_home: Path) -> Generator[None, None, None]:
    path_to_dags = airflow_home / "airflow_dags"
    with environ({"AIRFLOW_HOME": str(airflow_home)}):
        try:
            # Start by removing old cruft if it exists
            remove_airflow_home_remnants(airflow_home)
            # Scaffold the airflow configuration file.
            chmod_script(airflow_cfg_script_path())
            subprocess.run(
                [str(airflow_cfg_script_path()), str(path_to_dags), str(airflow_home)], check=False
            )
            yield
        finally:
            # Clean up after ourselves.
            remove_airflow_home_remnants(airflow_home)


def asset_spec(asset_str: str, defs: Definitions) -> Optional[AssetSpec]:
    """Get the spec of an asset from the definitions by its string representation."""
    return next(
        iter(
            spec
            for spec in defs.resolve_all_asset_specs()
            if spec.key.to_user_string() == asset_str
        ),
        None,
    )


def get_job_from_defs(
    name: str, defs: Definitions
) -> Optional[Union[JobDefinition, UnresolvedAssetJobDefinition]]:
    """Get the job from the definitions by its name."""
    return next(
        iter(job for job in (defs.jobs or []) if job.name == name),
        None,
    )


@dataclass
class BasicAssetComponent(Component, Resolvable):
    key: ResolvedAssetKey

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        @asset(key=self.key)
        def asset_def():
            pass

        return Definitions(assets=[asset_def])



import asyncio
from abc import ABC, abstractmethod
from collections.abc import Iterator
from itertools import chain
from typing import Optional, Union

from dagster import (
    AssetMaterialization,
    _check as check,
)
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.metadata.metadata_value import MetadataValue
from dagster._core.events import (
    AssetMaterializationPlannedData,
    DagsterEvent,
    DagsterEventType,
    JobFailureData,
    StepMaterializationData,
)
from dagster._core.execution.context.op_execution_context import OpExecutionContext
from dagster._core.remote_origin import RemoteJobOrigin
from dagster._core.storage.dagster_run import DagsterRun, DagsterRunStatus
from dagster._core.storage.tags import EXTERNAL_JOB_SOURCE_TAG_KEY
from dagster._core.utils import make_new_run_id
from dagster._record import record
from dagster._time import datetime_from_timestamp
from dagster._utils.merger import merge_dicts

from dagster_airlift.constants import (
    DAG_ID_TAG_KEY,
    DAG_RUN_ID_TAG_KEY,
    DAG_RUN_URL_TAG_KEY,
    NO_STEP_KEY,
    OBSERVATION_RUN_TAG_KEY,
)
from dagster_airlift.core.airflow_defs_data import AirflowDefinitionsData
from dagster_airlift.core.airflow_instance import AirflowInstance
from dagster_airlift.core.monitoring_job.utils import (
    extract_metadata_from_logs,
    get_dagster_run_for_airflow_repr,
    get_externally_managed_runs_from_handle,
    structured_log,
)
from dagster_airlift.core.runtime_representations import DagRun, TaskInstance


class AirflowEvent(ABC):
    @property
    @abstractmethod
    def timestamp(self) -> float:
        pass

    @abstractmethod
    def persist_state(
        self,
        context: OpExecutionContext,
        airflow_data: AirflowDefinitionsData,
        airflow_instance: AirflowInstance,
    ) -> None:
        """Persist the state of the event to Dagster."""
        pass


@record
class DagRunStarted(AirflowEvent):
    dag_run: DagRun

    @property
    def timestamp(self) -> float:
        return self.dag_run.start_date.timestamp()

    def persist_state(
        self,
        context: OpExecutionContext,
        airflow_data: AirflowDefinitionsData,
        airflow_instance: AirflowInstance,
    ) -> None:
        # If the run is already in Dagster, we don't need to do anything.
        relevant_job_def = airflow_data.airflow_mapped_jobs_by_dag_handle.get(
            self.dag_run.dag_handle
        )
        # Only if we have a relevant job def do we need to mess with creating runs.
        if not relevant_job_def:
            return
        dagster_run_id = make_new_run_id()
        context.instance.run_storage.add_historical_run(
            dagster_run=DagsterRun(
                run_id=dagster_run_id,
                job_name=relevant_job_def.name,
                tags={
                    DAG_RUN_ID_TAG_KEY: self.dag_run.run_id,
                    DAG_ID_TAG_KEY: self.dag_run.dag_id,
                    EXTERNAL_JOB_SOURCE_TAG_KEY: "airflow",
                    DAG_RUN_URL_TAG_KEY: self.dag_run.url,
                    OBSERVATION_RUN_TAG_KEY: "true",
                },
                status=DagsterRunStatus.NOT_STARTED,
                remote_job_origin=RemoteJobOrigin(
                    # We steal the repository origin from the original run.
                    # This allows the UI to link the run we're creating here back to the
                    # job in Dagster.
                    # It's not set in test contexts
                    repository_origin=context.run.remote_job_origin.repository_origin,
                    job_name=relevant_job_def.name,
                )
                if context.run.remote_job_origin
                else None,
            ),
            run_creation_time=self.dag_run.start_date,
        )

        context.instance.report_dagster_event(
            run_id=dagster_run_id,
            dagster_event=DagsterEvent(
                event_type_value="PIPELINE_START",
                job_name=relevant_job_def.name,
            ),
            timestamp=self.dag_run.start_date.timestamp(),
        )

        planned_asset_keys = {
            *airflow_data.all_asset_keys_by_dag_handle[self.dag_run.dag_handle],
            *{
                asset_key
                for task_handle, keys in airflow_data.mapped_asset_keys_by_task_handle.items()
                for asset_key in keys
                if task_handle.dag_id == self.dag_run.dag_id
            },
        }
        for asset_key in planned_asset_keys:
            context.instance.report_dagster_event(
                run_id=dagster_run_id,
                dagster_event=DagsterEvent.build_asset_materialization_planned_event(
                    job_name=relevant_job_def.name,
                    asset_materialization_planned_data=AssetMaterializationPlannedData(
                        asset_key=asset_key,
                    ),
                    step_key=NO_STEP_KEY,
                ),
                timestamp=self.dag_run.start_date.timestamp(),
            )


@record
class TaskInstanceCompleted(AirflowEvent):
    task_instance: TaskInstance
    metadata: dict[str, MetadataValue]

    @property
    def timestamp(self) -> float:
        return self.task_instance.end_date.timestamp()

    def persist_state(
        self,
        context: OpExecutionContext,
        airflow_data: AirflowDefinitionsData,
        airflow_instance: AirflowInstance,
    ) -> None:
        corresponding_run = get_dagster_run_for_airflow_repr(context, self.task_instance)
        externally_managed_runs = get_externally_managed_runs_from_handle(
            context, self.task_instance.task_handle, self.task_instance.run_id
        )

        per_key_metadata = (
            merge_dicts(
                *(_per_asset_metadata_from_run(context, run) for run in externally_managed_runs)
            )
            if externally_managed_runs
            else {}
        )

        for asset in airflow_data.mapped_asset_keys_by_task_handle[self.task_instance.task_handle]:
            # IMPROVEME: Add metadata to the materialization event.
            _report_materialization(
                context=context,
                corresponding_run=corresponding_run,
                materialization=AssetMaterialization(
                    asset_key=asset,
                    metadata=merge_dicts(per_key_metadata.get(asset, {}), self.metadata),
                ),
                airflow_event=self.task_instance,
            )


@record
class DagRunCompleted(AirflowEvent):
    dag_run: DagRun

    @property
    def timestamp(self) -> float:
        return self.dag_run.end_date.timestamp()

    def persist_state(
        self,
        context: OpExecutionContext,
        airflow_data: AirflowDefinitionsData,
        airflow_instance: AirflowInstance,
    ) -> None:
        corresponding_run = get_dagster_run_for_airflow_repr(context, self.dag_run)
        for asset in airflow_data.all_asset_keys_by_dag_handle[self.dag_run.dag_handle]:
            _report_materialization(
                context=context,
                corresponding_run=corresponding_run,
                materialization=AssetMaterialization(asset_key=asset),
                airflow_event=self.dag_run,
            )

        if not corresponding_run:
            context.log.warning(
                f"Airflow Monitoring Job: No corresponding Dagster run found for completed Airflow run {self.dag_run.run_id}. Skipping run event emission."
            )
            return
        dagster_run_id = corresponding_run.run_id
        if self.dag_run.state == "success":
            structured_log(
                context,
                f"Emitting pipeline success event for run {dagster_run_id}.",
            )
            event = DagsterEvent(
                event_type_value="PIPELINE_SUCCESS",
                job_name=corresponding_run.job_name,
            )
        else:
            structured_log(
                context,
                f"Emitting pipeline failure event for run {dagster_run_id}.",
            )
            event = DagsterEvent(
                event_type_value="PIPELINE_FAILURE",
                job_name=corresponding_run.job_name,
                event_specific_data=JobFailureData(
                    error=None, failure_reason=None, first_step_failure_event=None
                ),
            )
        context.instance.report_dagster_event(
            run_id=dagster_run_id, dagster_event=event, timestamp=self.dag_run.end_date.timestamp()
        )


def _process_started_runs(
    context: OpExecutionContext,
    airflow_data: AirflowDefinitionsData,
    airflow_instance: AirflowInstance,
    range_start: float,
    range_end: float,
) -> Iterator[DagRunStarted]:
    offset = 0
    while True:
        (newly_started_runs, total_entries) = airflow_instance.get_dag_runs_batch(
            states=["running", "queued"],
            # When monitoring started runs; we only need to monitor dags that have an associated job.
            dag_ids=[
                dag_handle.dag_id
                for dag_handle in airflow_data.airflow_mapped_jobs_by_dag_handle.keys()
            ],
            start_date_gte=datetime_from_timestamp(range_start),
            start_date_lte=datetime_from_timestamp(range_end),
            offset=offset,
        )
        structured_log(
            context,
            f"Found {len(newly_started_runs)} newly started runs in the time range {datetime_from_timestamp(range_start)} to {datetime_from_timestamp(range_end)}",
        )
        yield from (DagRunStarted(dag_run=dag_run) for dag_run in newly_started_runs)
        offset += len(newly_started_runs)
        if offset >= total_entries:
            break


def _process_completed_runs(
    context: OpExecutionContext,
    airflow_data: AirflowDefinitionsData,
    airflow_instance: AirflowInstance,
    range_start: float,
    range_end: float,
) -> Iterator[Union[DagRunStarted, DagRunCompleted]]:
    offset = 0
    dag_ids_to_query = airflow_data.dag_ids_with_mapped_asset_keys | {
        handle.dag_id for handle in airflow_data.airflow_mapped_jobs_by_dag_handle.keys()
    }
    while True:
        (newly_completed_runs, total_entries) = airflow_instance.get_dag_runs_batch(
            states=["success", "failed", "up_for_retry"],
            dag_ids=list(dag_ids_to_query),
            end_date_gte=datetime_from_timestamp(range_start),
            end_date_lte=datetime_from_timestamp(range_end),
            offset=offset,
        )
        structured_log(
            context,
            f"Found {len(newly_completed_runs)} newly completed runs in the time range {datetime_from_timestamp(range_start)} to {datetime_from_timestamp(range_end)}",
        )
        for run in newly_completed_runs:
            corresponding_dagster_run = get_dagster_run_for_airflow_repr(context, run)
            if not corresponding_dagster_run:
                yield DagRunStarted(dag_run=run)
            yield DagRunCompleted(dag_run=run)
        offset += len(newly_completed_runs)
        if offset >= total_entries:
            break


async def _retrieve_logs_for_task_instance(
    context: OpExecutionContext,
    airflow_instance: AirflowInstance,
    task_instance: TaskInstance,
) -> TaskInstanceCompleted:
    logs = await asyncio.to_thread(
        airflow_instance.get_task_instance_logs,
        task_instance.dag_id,
        task_instance.task_id,
        task_instance.run_id,
        task_instance.try_number,
    )
    try:
        metadata = extract_metadata_from_logs(context, logs)
    except Exception as e:
        context.log.warning(
            f"An unexpected error occurred while extracting metadata from logs: {e}. Skipping metadata extraction for task instance {task_instance.task_id}."
        )
        metadata = {}

    return TaskInstanceCompleted(task_instance=task_instance, metadata=metadata)


async def _async_process_task_instances(
    context: OpExecutionContext,
    airflow_instance: AirflowInstance,
    task_instances: list[TaskInstance],
) -> list[TaskInstanceCompleted]:
    results = await asyncio.gather(
        *(
            _retrieve_logs_for_task_instance(context, airflow_instance, task_instance)
            for task_instance in task_instances
        )
    )

    return results


def _process_task_instances(
    context: OpExecutionContext,
    airflow_data: AirflowDefinitionsData,
    airflow_instance: AirflowInstance,
    range_start: float,
    range_end: float,
) -> Iterator[TaskInstanceCompleted]:
    # You'll notice that the query pattern here is different from the dag run query pattern.
    # That's because early Airflow 2 versions don't support batch size queries for task instances.
    # NOTE: In the case where we have a run which completed between iterations, we'll be emitting events for the run _after_ we've marked the run as finished.
    # so the timeline will be a bit screwed up.
    task_instances = airflow_instance.get_task_instance_batch_time_range(
        states=["success"],
        dag_ids=list(airflow_data.dag_ids_with_mapped_asset_keys),
        end_date_gte=datetime_from_timestamp(range_start),
        end_date_lte=datetime_from_timestamp(range_end),
    )
    structured_log(
        context,
        f"Found {len(task_instances)} completed task instances in the time range {datetime_from_timestamp(range_start)} to {datetime_from_timestamp(range_end)}",
    )
    yield from asyncio.run(_async_process_task_instances(context, airflow_instance, task_instances))


def persist_events(
    context: OpExecutionContext,
    airflow_data: AirflowDefinitionsData,
    airflow_instance: AirflowInstance,
    range_start: float,
    range_end: float,
) -> None:
    for event in sorted(
        chain(
            _process_started_runs(context, airflow_data, airflow_instance, range_start, range_end),
            _process_completed_runs(
                context, airflow_data, airflow_instance, range_start, range_end
            ),
            _process_task_instances(
                context, airflow_data, airflow_instance, range_start, range_end
            ),
        ),
        key=lambda event: event.timestamp,
    ):
        event.persist_state(context, airflow_data, airflow_instance)


def _report_materialization(
    *,
    context: OpExecutionContext,
    corresponding_run: Optional[DagsterRun],
    materialization: AssetMaterialization,
    airflow_event: Union[TaskInstance, DagRun],
) -> None:
    if corresponding_run:
        context.instance.report_dagster_event(
            run_id=corresponding_run.run_id,
            dagster_event=DagsterEvent(
                event_type_value="ASSET_MATERIALIZATION",
                job_name=corresponding_run.job_name,
                event_specific_data=StepMaterializationData(materialization=materialization),
            ),
            timestamp=airflow_event.end_date.timestamp(),
        )
    else:
        # Could also support timestamp override here; but would only benefit jobless Airlift.
        context.instance.report_runless_asset_event(
            asset_event=materialization,
        )


def _get_output_to_asset_key_map(
    context: OpExecutionContext, run: DagsterRun
) -> dict[AssetKey, str]:
    execution_plan_snap = context.instance.get_execution_plan_snapshot(
        check.not_none(run.execution_plan_snapshot_id)
    )
    mapping = {}
    for step in execution_plan_snap.steps:
        for output in step.outputs:
            props = check.not_none(output.properties)
            if props.asset_key:
                mapping[output.name] = props.asset_key
    return mapping


def _per_asset_metadata_from_run(
    context: OpExecutionContext, run: DagsterRun
) -> dict[AssetKey, dict[str, MetadataValue]]:
    output_to_key_map = _get_output_to_asset_key_map(context, run)
    conn = context.instance.event_log_storage.get_records_for_run(
        run_id=run.run_id,
        of_type=DagsterEventType.STEP_OUTPUT,
    )
    per_key_metadata = {}
    for event_record in conn.records:
        event = check.not_none(event_record.event_log_entry.dagster_event)
        if event.step_output_data.output_name in output_to_key_map:
            asset_key = output_to_key_map[event.step_output_data.output_name]
            if event.step_output_data.metadata:
                per_key_metadata[asset_key] = event.step_output_data.metadata
    return per_key_metadata



import json
import os
from abc import abstractmethod
from collections.abc import Iterable, Mapping, Sequence
from typing import Any

import requests
from airflow import DAG

from dagster_airlift.constants import (
    DAG_MAPPING_METADATA_KEY,
    DEFER_ASSET_EVENTS_TAG,
    TASK_ID_TAG_KEY,
)
from dagster_airlift.in_airflow.base_asset_operator import BaseDagsterAssetsOperator, Context


class BaseProxyDAGToDagsterOperator(BaseDagsterAssetsOperator):
    """An operator base class that proxies the entire DAG's execution to Dagster assets with
    metadata that map to the DAG id used by this task.

    For the Dag ID that this operator proxies, it expects there to be corresponding assets
    in the linked Dagster deployment that have metadata entries with the key `dagster-airlift/dag-mapping` that
    map to this Dag ID. This metadata is typically set using the
    :py:func:`dagster_airlift.core.assets_with_dag_mappings` function.

    The following methods must be implemented by subclasses:

        - :py:meth:`get_dagster_session` (inherited from :py:class:`BaseDagsterAssetsOperator`)
        - :py:meth:`get_dagster_url` (inherited from :py:class:`BaseDagsterAssetsOperator`)
        - :py:meth:`build_from_dag` A class method which takes the DAG to be proxied, and constructs
            an instance of this operator from it.

    There is a default implementation of this operator, :py:class:`DefaultProxyDAGToDagsterOperator`,
    which is used by :py:func:`proxying_to_dagster` if no override operator is provided.
    """

    def filter_asset_nodes(
        self, context: Context, asset_nodes: Sequence[Mapping[str, Any]]
    ) -> Iterable[Mapping[str, Any]]:
        for asset_node in asset_nodes:
            if matched_dag_id(asset_node, self.get_airflow_dag_id(context)):
                yield asset_node

    @property
    def should_defer_asset_events(self) -> bool:
        return False

    def default_dagster_run_tags(self, context: Mapping[str, Any]) -> dict[str, str]:
        tags = dict(super().default_dagster_run_tags(context))
        if self.should_defer_asset_events:
            tags[DEFER_ASSET_EVENTS_TAG] = "true"
        tags[TASK_ID_TAG_KEY] = "FULL_DAG_PROXY"
        return tags

    @classmethod
    @abstractmethod
    def build_from_dag(cls, dag: DAG) -> "BaseProxyDAGToDagsterOperator":
        """Builds a proxy operator from the passed-in DAG."""


class DefaultProxyDAGToDagsterOperator(BaseProxyDAGToDagsterOperator):
    """The default task proxying operator - which opens a blank session and expects the dagster URL to be set in the environment.
    The dagster url is expected to be set in the environment as DAGSTER_URL.

    This operator should not be instantiated directly - it is instantiated by :py:func:`proxying_to_dagster` if no
    override operator is provided.
    """

    def get_dagster_session(self, context: Context) -> requests.Session:
        return requests.Session()

    def get_dagster_url(self, context: Context) -> str:
        return os.environ["DAGSTER_URL"]

    @classmethod
    def build_from_dag(cls, dag: DAG) -> "DefaultProxyDAGToDagsterOperator":
        return DefaultProxyDAGToDagsterOperator(
            task_id=f"DAGSTER_OVERRIDE_DAG_{dag.dag_id}",
            dag=dag,
        )


def matched_dag_id(asset_node: Mapping[str, Any], dag_id: str) -> bool:
    json_metadata_entries = {
        entry["label"]: entry["jsonString"]
        for entry in asset_node["metadataEntries"]
        if entry["__typename"] == "JsonMetadataEntry"
    }

    mapping_entry = json_metadata_entries.get(DAG_MAPPING_METADATA_KEY)
    if mapping_entry:
        mappings = json.loads(mapping_entry)
        return any(mapping["dag_id"] == dag_id for mapping in mappings)
    return False



from collections.abc import Mapping
from typing import Any, Optional

from dagster import AssetKey, JsonMetadataValue, MarkdownMetadataValue
from dagster._core.definitions.metadata.metadata_value import UrlMetadataValue

from dagster_airlift.constants import PEERED_DAG_MAPPING_METADATA_KEY, SOURCE_CODE_METADATA_KEY
from dagster_airlift.core.airflow_instance import DagInfo


def dag_description(dag_info: DagInfo) -> str:
    return f"""
    A materialization corresponds to a successful run of airflow DAG {dag_info.dag_id}.
    """


def dag_asset_metadata(dag_info: DagInfo) -> dict[str, Any]:
    return {
        "Dag Info (raw)": JsonMetadataValue(dag_info.metadata),
        "Dag ID": dag_info.dag_id,
        "Link to DAG": UrlMetadataValue(dag_info.url),
    }


def peered_dag_asset_metadata(dag_info: DagInfo, source_code: Optional[str]) -> Mapping[str, Any]:
    metadata = dag_asset_metadata(dag_info)
    metadata[PEERED_DAG_MAPPING_METADATA_KEY] = [{"dag_id": dag_info.dag_id}]
    if source_code:
        metadata[SOURCE_CODE_METADATA_KEY] = MarkdownMetadataValue(
            f"""
```python
{source_code}
```
"""
        )
    return metadata


def get_leaf_assets_for_dag(
    asset_keys_in_dag: set[AssetKey],
    downstreams_asset_dependency_graph: dict[AssetKey, set[AssetKey]],
) -> set[AssetKey]:
    # An asset is a "leaf" for the dag if it has no transitive dependencies _within_ the dag. It may have
    # dependencies _outside_ the dag.
    leaf_assets = []
    cache = {}
    for asset_key in asset_keys_in_dag:
        if (
            get_transitive_dependencies_for_asset(
                asset_key, downstreams_asset_dependency_graph, cache
            ).intersection(asset_keys_in_dag)
            == set()
        ):
            leaf_assets.append(asset_key)
    return set(leaf_assets)


def get_transitive_dependencies_for_asset(
    asset_key: AssetKey,
    downstreams_asset_dependency_graph: dict[AssetKey, set[AssetKey]],
    cache: dict[AssetKey, set[AssetKey]],
) -> set[AssetKey]:
    if asset_key in cache:
        return cache[asset_key]
    transitive_deps = set()
    for dep in downstreams_asset_dependency_graph[asset_key]:
        transitive_deps.add(dep)
        transitive_deps.update(
            get_transitive_dependencies_for_asset(dep, downstreams_asset_dependency_graph, cache)
        )
    cache[asset_key] = transitive_deps
    return transitive_deps



PEERED_DAG_MAPPING_METADATA_KEY = "dagster-airlift/peered-dag-mapping"
DAG_MAPPING_METADATA_KEY = "dagster-airlift/dag-mapping"
AIRFLOW_SOURCE_METADATA_KEY_PREFIX = "dagster-airlift/source"
TASK_MAPPING_METADATA_KEY = "dagster-airlift/task-mapping"
AUTOMAPPED_TASK_METADATA_KEY = "dagster-airlift/automapped-task"
# This represents the timestamp used in ordering the materializatons.
EFFECTIVE_TIMESTAMP_METADATA_KEY = "dagster-airlift/effective-timestamp"
AIRFLOW_TASK_INSTANCE_LOGICAL_DATE_METADATA_KEY = (
    "dagster-airlift/airflow-task-instance-logical-date"
)
AIRFLOW_RUN_ID_METADATA_KEY = "dagster-airlift/airflow-run-id"
DAG_RUN_ID_TAG_KEY = "dagster-airlift/airflow-dag-run-id"
DAG_ID_TAG_KEY = "dagster-airlift/airflow-dag-id"
TASK_ID_TAG_KEY = "dagster-airlift/airflow-task-id"
DAG_RUN_URL_TAG_KEY = "dagster-airlift/airflow-dag-run-url"
OBSERVATION_RUN_TAG_KEY = "dagster-airlift/observation-run"

DEFER_ASSET_EVENTS_TAG = "dagster/defer_asset_events"

SOURCE_CODE_METADATA_KEY = "Source Code"

NO_STEP_KEY = "__no_step__"



from collections.abc import Sequence
from typing import Optional

from dagster._record import record


@record
class AirflowFilter:
    """Filters the set of Airflow objects to fetch.

    Args:
        dag_id_ilike (Optional[str]): A pattern used to match the set of dag_ids to retrieve. Uses the sql ILIKE operator Airflow-side.
        airflow_tags (Optional[Sequence[str]]): Filters down to the set of Airflow DAGs whcih contain the particular tags provided.
        retrieve_datasets (bool): Whether to retrieve datasets from Airflow. Defaults to True.
        dataset_uri_ilike (Optional[str]): A pattern used to match the set of datasets to retrieve. Uses the sql ILIKE operator Airflow-side.
    """

    dag_id_ilike: Optional[str] = None
    airflow_tags: Optional[Sequence[str]] = None
    retrieve_datasets: bool = True
    dataset_uri_ilike: Optional[str] = None

    def augment_request_params(self, request_params: dict) -> dict:
        new_request_params = request_params.copy()
        if self.dag_id_ilike is not None:
            new_request_params["dag_id_pattern"] = self.dag_id_ilike
        if self.airflow_tags is not None:
            new_request_params["tags"] = self.airflow_tags
        return new_request_params



import datetime
import time
from abc import ABC
from collections.abc import Sequence
from typing import Any, Optional

import requests
from dagster._annotations import beta, public
from dagster._core.definitions.utils import check_valid_name
from dagster._time import get_current_datetime
from dagster_shared.error import DagsterError

from dagster_airlift.core.filter import AirflowFilter
from dagster_airlift.core.runtime_representations import DagRun, TaskInstance
from dagster_airlift.core.serialization.serialized_data import (
    DagInfo,
    Dataset,
    DatasetConsumingDag,
    DatasetProducingTask,
    TaskInfo,
)

TERMINAL_STATES = {"success", "failed", "canceled"}
# This limits the number of task ids that we attempt to query from airflow's task instance rest API at a given time.
# Airflow's batch task instance retrieval rest API doesn't have a limit parameter, but we query a single run at a time, meaning we should be getting
# a single task instance per task id.
# Airflow task instance batch API: https://airflow.apache.org/docs/apache-airflow/stable/stable-rest-api-ref.html#operation/get_task_instances_batch
DEFAULT_BATCH_TASK_RETRIEVAL_LIMIT = 100
# This corresponds directly to the page_limit parameter on airflow's batch dag runs rest API.
# Airflow dag run batch API: https://airflow.apache.org/docs/apache-airflow/stable/stable-rest-api-ref.html#operation/get_dag_runs_batch
DEFAULT_BATCH_DAG_RUNS_LIMIT = 100
DEFAULT_DAG_LIST_LIMIT = 100
SLEEP_SECONDS = 1


@beta
class AirflowAuthBackend(ABC):
    """An abstract class that represents an authentication backend for an Airflow instance.

    Requires two methods to be implemented by subclasses:
    - get_session: Returns a requests.Session object that can be used to make requests to the Airflow instance, and handles authentication.
    - get_webserver_url: Returns the base URL of the Airflow webserver.

    The `dagster-airlift` package provides the following default implementations:
    - :py:class:`dagster-airlift.core.AirflowBasicAuthBackend`: An authentication backend that uses Airflow's basic auth to authenticate with the Airflow instance.
    - :py:class:`dagster-airlift.mwaa.MwaaSessionAuthBackend`: An authentication backend that uses AWS MWAA's web login token to authenticate with the Airflow instance (requires `dagster-airlift[mwaa]`).
    """

    def get_session(self) -> requests.Session:
        raise NotImplementedError("This method must be implemented by subclasses.")

    def get_webserver_url(self) -> str:
        raise NotImplementedError("This method must be implemented by subclasses.")


@beta
class AirflowInstance:
    """A class that represents a running Airflow Instance and provides methods for interacting with its REST API.

    Args:
        auth_backend (AirflowAuthBackend): The authentication backend to use when making requests to the Airflow instance.
        name (str): The name of the Airflow instance. This will be prefixed to any assets automatically created using this instance.
        batch_task_instance_limit (int): The number of task instances to query at a time when fetching task instances. Defaults to 100.
        batch_dag_runs_limit (int): The number of dag runs to query at a time when fetching dag runs. Defaults to 100.
    """

    def __init__(
        self,
        auth_backend: AirflowAuthBackend,
        name: str,
        batch_task_instance_limit: int = DEFAULT_BATCH_TASK_RETRIEVAL_LIMIT,
        batch_dag_runs_limit: int = DEFAULT_BATCH_DAG_RUNS_LIMIT,
        dag_list_limit: int = DEFAULT_DAG_LIST_LIMIT,
    ) -> None:
        self.auth_backend = auth_backend
        self.name = check_valid_name(name)
        self.batch_task_instance_limit = batch_task_instance_limit
        self.batch_dag_runs_limit = batch_dag_runs_limit
        self.dag_list_limit = dag_list_limit

    @property
    def normalized_name(self) -> str:
        return self.name.replace(" ", "_").replace("-", "_")

    def get_api_url(self) -> str:
        return f"{self.auth_backend.get_webserver_url()}/api/v1"

    def list_dags(self, retrieval_filter: Optional[AirflowFilter] = None) -> list["DagInfo"]:
        retrieval_filter = retrieval_filter or AirflowFilter()
        dag_responses = []
        webserver_url = self.auth_backend.get_webserver_url()

        while True:
            params = retrieval_filter.augment_request_params(
                {
                    "limit": self.dag_list_limit,
                    "offset": len(dag_responses),
                }
            )
            response = self.auth_backend.get_session().get(
                f"{self.get_api_url()}/dags",
                params=params,
            )
            if response.status_code == 200:
                dags = response.json()
                dag_responses.extend(
                    DagInfo(
                        webserver_url=webserver_url,
                        dag_id=dag["dag_id"],
                        metadata=dag,
                    )
                    for dag in dags["dags"]
                )
                if len(dags["dags"]) < self.dag_list_limit:
                    break
            else:
                raise DagsterError(
                    f"Failed to fetch DAGs. Status code: {response.status_code}, Message: {response.text}"
                )
        return dag_responses

    def list_variables(self) -> list[dict[str, Any]]:
        response = self.auth_backend.get_session().get(f"{self.get_api_url()}/variables")
        if response.status_code == 200:
            return response.json()["variables"]
        else:
            raise DagsterError(
                "Failed to fetch variables. Status code: {response.status_code}, Message: {response.text}"
            )

    def get_task_instance_batch_time_range(
        self,
        dag_ids: Sequence[str],
        states: Sequence[str],
        end_date_gte: datetime.datetime,
        end_date_lte: datetime.datetime,
    ) -> list["TaskInstance"]:
        """Get all task instances across all dag_ids for a given time range."""
        response = self.auth_backend.get_session().post(
            f"{self.get_api_url()}/dags/~/dagRuns/~/taskInstances/list",
            json={
                "dag_ids": dag_ids,
                "end_date_gte": end_date_gte.isoformat(),
                "end_date_lte": end_date_lte.isoformat(),
                # Airflow's API refers to this variable in the singular, but it's actually a list. We keep the confusion contained to this one function.
                "state": states,
            },
        )

        if response.status_code != 200:
            raise DagsterError(
                f"Failed to fetch task instances for {dag_ids}. Status code: {response.status_code}, Message: {response.text}"
            )
        return [
            TaskInstance(
                webserver_url=self.auth_backend.get_webserver_url(),
                dag_id=task_instance_json["dag_id"],
                task_id=task_instance_json["task_id"],
                run_id=task_instance_json["dag_run_id"],
                metadata=task_instance_json,
            )
            for task_instance_json in response.json()["task_instances"]
        ]

    def get_task_instance_batch(
        self, dag_id: str, task_ids: Sequence[str], run_id: str, states: Sequence[str]
    ) -> list["TaskInstance"]:
        """Get all task instances for a given dag_id, task_ids, and run_id."""
        # It's not possible to offset the task instance API on versions of Airflow < 2.7.0, so we need to
        # chunk the task ids directly.
        task_instances = []
        task_id_chunks = [
            task_ids[i : i + self.batch_task_instance_limit]
            for i in range(0, len(task_ids), self.batch_task_instance_limit)
        ]
        for task_id_chunk in task_id_chunks:
            response = self.auth_backend.get_session().post(
                f"{self.get_api_url()}/dags/~/dagRuns/~/taskInstances/list",
                json={
                    "dag_ids": [dag_id],
                    "task_ids": task_id_chunk,
                    "dag_run_ids": [run_id],
                },
            )

            if response.status_code == 200:
                for task_instance_json in response.json()["task_instances"]:
                    task_id = task_instance_json["task_id"]
                    task_instance = TaskInstance(
                        webserver_url=self.auth_backend.get_webserver_url(),
                        dag_id=dag_id,
                        task_id=task_id,
                        run_id=run_id,
                        metadata=task_instance_json,
                    )
                    if task_instance.state in states:
                        task_instances.append(task_instance)
            else:
                raise DagsterError(
                    f"Failed to fetch task instances for {dag_id}/{task_id_chunk}/{run_id}. Status code: {response.status_code}, Message: {response.text}"
                )
        return task_instances

    def get_task_instance(self, dag_id: str, task_id: str, run_id: str) -> "TaskInstance":
        response = self.auth_backend.get_session().get(
            f"{self.get_api_url()}/dags/{dag_id}/dagRuns/{run_id}/taskInstances/{task_id}"
        )
        if response.status_code == 200:
            return TaskInstance(
                webserver_url=self.auth_backend.get_webserver_url(),
                dag_id=dag_id,
                task_id=task_id,
                run_id=run_id,
                metadata=response.json(),
            )
        else:
            raise DagsterError(
                f"Failed to fetch task instance for {dag_id}/{task_id}/{run_id}. Status code: {response.status_code}, Message: {response.text}"
            )

    def get_task_infos(self, *, dag_id: str) -> list["TaskInfo"]:
        response = self.auth_backend.get_session().get(f"{self.get_api_url()}/dags/{dag_id}/tasks")

        if response.status_code != 200:
            raise DagsterError(
                f"Failed to fetch task infos for {dag_id}. Status code: {response.status_code}, Message: {response.text}"
            )

        dag_json = response.json()
        webserver_url = self.auth_backend.get_webserver_url()
        return [
            TaskInfo(
                webserver_url=webserver_url,
                dag_id=dag_id,
                metadata=task_data,
                task_id=task_data["task_id"],
            )
            for task_data in dag_json["tasks"]
        ]

    def get_task_info(self, *, dag_id: str, task_id: str) -> "TaskInfo":
        response = self.auth_backend.get_session().get(
            f"{self.get_api_url()}/dags/{dag_id}/tasks/{task_id}"
        )
        if response.status_code == 200:
            return TaskInfo(
                webserver_url=self.auth_backend.get_webserver_url(),
                dag_id=dag_id,
                task_id=task_id,
                metadata=response.json(),
            )
        else:
            raise DagsterError(
                f"Failed to fetch task info for {dag_id}/{task_id}. Status code: {response.status_code}, Message: {response.text}"
            )

    def get_dag_source_code(self, file_token: str) -> str:
        response = self.auth_backend.get_session().get(
            f"{self.get_api_url()}/dagSources/{file_token}"
        )
        if response.status_code == 200:
            return response.text
        else:
            raise DagsterError(
                f"Failed to fetch source code. Status code: {response.status_code}, Message: {response.text}"
            )

    def get_dag_runs(
        self, dag_id: str, start_date: datetime.datetime, end_date: datetime.datetime
    ) -> list["DagRun"]:
        response = self.auth_backend.get_session().get(
            f"{self.get_api_url()}/dags/{dag_id}/dagRuns",
            params={
                "updated_at_gte": start_date.isoformat(),
                "updated_at_lte": end_date.isoformat(),
                "state": ["success"],
            },
        )
        if response.status_code == 200:
            webserver_url = self.auth_backend.get_webserver_url()
            return [
                DagRun(
                    webserver_url=webserver_url,
                    dag_id=dag_id,
                    run_id=dag_run["dag_run_id"],
                    metadata=dag_run,
                )
                for dag_run in response.json()["dag_runs"]
            ]
        else:
            raise DagsterError(
                f"Failed to fetch dag runs for {dag_id}. Status code: {response.status_code}, Message: {response.text}"
            )

    def get_dag_runs_batch(
        self,
        dag_ids: Sequence[str],
        end_date_gte: Optional[datetime.datetime] = None,
        end_date_lte: Optional[datetime.datetime] = None,
        start_date_gte: Optional[datetime.datetime] = None,
        start_date_lte: Optional[datetime.datetime] = None,
        offset: int = 0,
        states: Optional[Sequence[str]] = None,
    ) -> tuple[list["DagRun"], int]:
        """For the given list of dag_ids, return a tuple containing:
        - A list of dag runs ending within (end_date_gte, end_date_lte). Returns a maximum of batch_dag_runs_limit (which is configurable on the instance).
        - The number of total rows returned.
        """
        states = states or ["success"]
        params = {
            "dag_ids": list(dag_ids),
            "order_by": "end_date",
            "states": states,
            "page_offset": offset,
            "page_limit": self.batch_dag_runs_limit,
        }
        if end_date_gte:
            params["end_date_gte"] = end_date_gte.isoformat()
        if end_date_lte:
            params["end_date_lte"] = end_date_lte.isoformat()
        if start_date_gte:
            params["start_date_gte"] = start_date_gte.isoformat()
        if start_date_lte:
            params["start_date_lte"] = start_date_lte.isoformat()

        response = self.auth_backend.get_session().post(
            f"{self.get_api_url()}/dags/~/dagRuns/list",
            json=params,
        )
        if response.status_code == 200:
            webserver_url = self.auth_backend.get_webserver_url()
            return (
                [
                    DagRun(
                        webserver_url=webserver_url,
                        dag_id=dag_run["dag_id"],
                        run_id=dag_run["dag_run_id"],
                        metadata=dag_run,
                    )
                    for dag_run in response.json()["dag_runs"]
                ],
                response.json()["total_entries"],
            )
        else:
            raise DagsterError(
                f"Failed to fetch dag runs for {dag_ids}. Status code: {response.status_code}, Message: {response.text}"
            )

    @public
    def trigger_dag(self, dag_id: str, logical_date: Optional[datetime.datetime] = None) -> str:
        """Trigger a dag run for the given dag_id.

        Does not wait for the run to finish. To wait for the completed run to finish, use :py:meth:`wait_for_run_completion`.

        Args:
            dag_id (str): The dag id to trigger.
            logical_date (Optional[datetime.datetime]): The Airflow logical_date to use for the dag run. If not provided, the current time will be used. Previously known as execution_date in Airflow; find more information in the Airflow docs: https://airflow.apache.org/docs/apache-airflow/stable/faq.html#what-does-execution-date-mean

        Returns:
            str: The dag run id.
        """
        params = {} if not logical_date else {"logical_date": logical_date.isoformat()}
        response = self.auth_backend.get_session().post(
            f"{self.get_api_url()}/dags/{dag_id}/dagRuns",
            json=params,
        )
        if response.status_code != 200:
            raise DagsterError(
                f"Failed to launch run for {dag_id}. Status code: {response.status_code}, Message: {response.text}"
            )
        return response.json()["dag_run_id"]

    def get_dag_run(self, dag_id: str, run_id: str) -> "DagRun":
        response = self.auth_backend.get_session().get(
            f"{self.get_api_url()}/dags/{dag_id}/dagRuns/{run_id}"
        )
        if response.status_code != 200:
            raise DagsterError(
                f"Failed to fetch dag run for {dag_id}/{run_id}. Status code: {response.status_code}, Message: {response.text}"
            )
        return DagRun(
            webserver_url=self.auth_backend.get_webserver_url(),
            dag_id=dag_id,
            run_id=run_id,
            metadata=response.json(),
        )

    def unpause_dag(self, dag_id: str) -> None:
        response = self.auth_backend.get_session().patch(
            f"{self.get_api_url()}/dags",
            json={"is_paused": False},
            params={"dag_id_pattern": dag_id},
        )
        if response.status_code != 200:
            raise DagsterError(
                f"Failed to unpause dag {dag_id}. Status code: {response.status_code}, Message: {response.text}"
            )

    @public
    def wait_for_run_completion(self, dag_id: str, run_id: str, timeout: int = 30) -> None:
        """Given a run ID of an airflow dag, wait for that run to reach a completed state.

        Args:
            dag_id (str): The dag id.
            run_id (str): The run id.
            timeout (int): The number of seconds to wait before timing out.

        Returns:
            None
        """
        start_time = get_current_datetime()
        while get_current_datetime() - start_time < datetime.timedelta(seconds=timeout):
            dag_run = self.get_dag_run(dag_id, run_id)
            if dag_run.finished:
                return
            time.sleep(
                SLEEP_SECONDS
            )  # Sleep for a second before checking again. This way we don't flood the rest API with requests.
        raise DagsterError(f"Timed out waiting for airflow run {run_id} to finish.")

    @public
    def get_run_state(self, dag_id: str, run_id: str) -> str:
        """Given a run ID of an airflow dag, return the state of that run.

        Args:
            dag_id (str): The dag id.
            run_id (str): The run id.

        Returns:
            str: The state of the run. Will be one of the states defined by Airflow.
        """
        return self.get_dag_run(dag_id, run_id).state

    def delete_run(self, dag_id: str, run_id: str) -> None:
        response = self.auth_backend.get_session().delete(
            f"{self.get_api_url()}/dags/{dag_id}/dagRuns/{run_id}"
        )
        if response.status_code != 204:
            raise DagsterError(
                f"Failed to delete run for {dag_id}/{run_id}. Status code: {response.status_code}, Message: {response.text}"
            )
        return None

    def _get_datasets(
        self,
        *,
        limit: int,
        offset: int,
        retrieval_filter: AirflowFilter,
        dag_ids: Optional[Sequence[str]],
    ) -> Sequence["Dataset"]:
        params: dict[str, Any] = {"limit": limit, "offset": offset}
        if retrieval_filter.dataset_uri_ilike:
            params["uri_pattern"] = retrieval_filter.dataset_uri_ilike

        response = self.auth_backend.get_session().get(
            f"{self.get_api_url()}/datasets",
            params=params,
        )

        if response.status_code != 200:
            raise DagsterError(
                f"Failed to fetch datasets. Status code: {response.status_code}, Message: {response.text}"
            )

        data = response.json()
        datasets = []

        for dataset_data in data["datasets"]:
            producing_tasks = [
                DatasetProducingTask(
                    dag_id=task_data["dag_id"],
                    task_id=task_data["task_id"],
                    created_at=task_data.get("created_at", ""),
                    updated_at=task_data.get("updated_at", ""),
                )
                for task_data in dataset_data.get("producing_tasks", [])
                if not dag_ids or task_data["dag_id"] in dag_ids
            ]
            if not producing_tasks:
                # If the dataset has no producers among the set of dag_ids we care about, skip it.
                continue

            consuming_dags = [
                DatasetConsumingDag(
                    dag_id=dag_data["dag_id"],
                    created_at=dag_data.get("created_at", ""),
                    updated_at=dag_data.get("updated_at", ""),
                )
                for dag_data in dataset_data.get("consuming_dags", [])
                # Skip consuming dags that are not in the set of dag_ids we care about.
                if not dag_ids or dag_data["dag_id"] in dag_ids
            ]

            dataset = Dataset(
                id=dataset_data["id"],
                uri=dataset_data["uri"],
                extra=dataset_data.get("extra", {}),
                created_at=dataset_data.get("created_at", ""),
                updated_at=dataset_data.get("updated_at", ""),
                consuming_dags=consuming_dags,
                producing_tasks=producing_tasks,
            )

            datasets.append(dataset)

        return datasets

    def get_all_datasets(
        self,
        *,
        batch_size: int = 100,
        retrieval_filter: Optional[AirflowFilter] = None,
        dag_ids: Optional[Sequence[str]] = None,
    ) -> Sequence["Dataset"]:
        """Get all datasets from the Airflow instance, handling pagination.

        Args:
            batch_size: The number of items to fetch per request. Default is 100.
            retrieval_filter: An optional filter to apply to the dataset retrieval.
            dag_ids: One or more DAG IDs to filter datasets by associated DAGs either consuming or producing.

        Returns:
            A list of Dataset objects.
        """
        datasets = []
        offset = 0
        retrieval_filter = retrieval_filter or AirflowFilter()

        while True:
            batch = self._get_datasets(
                limit=batch_size,
                offset=offset,
                retrieval_filter=retrieval_filter,
                dag_ids=dag_ids,
            )

            datasets.extend(batch)

            if len(batch) < batch_size:  # No more datasets to fetch
                break

            offset += batch_size

        return datasets

    def get_task_instance_logs(
        self, dag_id: str, task_id: str, run_id: str, try_number: int
    ) -> str:
        continuation_token = None
        logs = []
        while True:
            response = self.auth_backend.get_session().get(
                f"{self.get_api_url()}/dags/{dag_id}/dagRuns/{run_id}/taskInstances/{task_id}/logs/{try_number}",
                headers={"Accept": "application/json"},
                params={"token": continuation_token} if continuation_token else None,
                timeout=5,
            )
            if response.status_code != 200:
                raise DagsterError(
                    f"Failed to fetch task instance logs for {dag_id}/{task_id}/{run_id}/{try_number}. Status code: {response.status_code}, Message: {response.text}"
                )
            data = response.json()
            # Love how it's different in the two cases lol.
            continuation_token = data.get("continuation_token")
            log = parse_af_log_response(data["content"])
            logs.append(log)
            if not continuation_token or log == "":
                break
        return "".join(logs)


def parse_af_log_response(logs: str) -> str:
    import ast

    parsed_data: list = ast.literal_eval(logs)
    strs = []
    for log_item in parsed_data:
        strs.append(log_item[1])
    return "".join(strs)



from collections import defaultdict
from collections.abc import Mapping, Sequence
from functools import cached_property
from typing import AbstractSet, Any, NamedTuple, Optional  # noqa: UP035

from dagster import (
    AssetKey,
    _check as check,
)
from dagster._annotations import PublicAttr, beta
from dagster._record import record
from dagster._serdes import whitelist_for_serdes

from dagster_airlift.constants import (
    DAG_ID_TAG_KEY,
    DAG_MAPPING_METADATA_KEY,
    TASK_ID_TAG_KEY,
    TASK_MAPPING_METADATA_KEY,
)


###################################################################################################
# Serialized data that represents airflow DAGs and tasks.
###################################################################################################
@whitelist_for_serdes
@record
class TaskInfo:
    webserver_url: str
    dag_id: str
    task_id: str
    metadata: dict[str, Any]

    @property
    def dag_url(self) -> str:
        return f"{self.webserver_url}/dags/{self.dag_id}"

    @cached_property
    def downstream_task_ids(self) -> list[str]:
        return check.is_list(self.metadata["downstream_task_ids"], str)


@beta
@whitelist_for_serdes
@record
class DagInfo:
    """A record containing information about a given airflow dag.

    Users should not instantiate this class directly. It is provided when customizing which DAGs are included
    in the generated definitions using the `dag_selector_fn` argument of :py:func:`build_defs_from_airflow_instance`.

    Args:
        metadata (Dict[str, Any]): The metadata associated with the dag, retrieved by the Airflow REST API:
            https://airflow.apache.org/docs/apache-airflow/stable/stable-rest-api-ref.html#operation/get_dags
    """

    webserver_url: str
    dag_id: str
    metadata: PublicAttr[dict[str, Any]]

    @property
    def url(self) -> str:
        return f"{self.webserver_url}/dags/{self.dag_id}"

    @property
    def file_token(self) -> str:
        return self.metadata["file_token"]


@whitelist_for_serdes
class TaskHandle(NamedTuple):
    dag_id: str
    task_id: str

    @property
    def metadata_key(self) -> str:
        return TASK_MAPPING_METADATA_KEY

    @property
    def as_dict(self) -> Mapping[str, Any]:
        return {
            "dag_id": self.dag_id,
            "task_id": self.task_id,
        }

    @property
    def identifying_tags(self) -> dict[str, str]:
        return {
            DAG_ID_TAG_KEY: self.dag_id,
            TASK_ID_TAG_KEY: self.task_id,
        }


@whitelist_for_serdes
class DagHandle(NamedTuple):
    dag_id: str

    @property
    def metadata_key(self) -> str:
        return DAG_MAPPING_METADATA_KEY

    @property
    def as_dict(self) -> Mapping[str, Any]:
        return {
            "dag_id": self.dag_id,
        }

    @property
    def identifying_tags(self) -> dict[str, str]:
        return {
            DAG_ID_TAG_KEY: self.dag_id,
        }


@whitelist_for_serdes
class DatasetConsumingDag(NamedTuple):
    dag_id: str
    created_at: str
    updated_at: str


@whitelist_for_serdes
class DatasetProducingTask(NamedTuple):
    dag_id: str
    task_id: str
    created_at: str
    updated_at: str


@whitelist_for_serdes
class Dataset(NamedTuple):
    id: int
    uri: str
    extra: Mapping[str, Any]
    created_at: str
    updated_at: str
    consuming_dags: Sequence[DatasetConsumingDag]
    producing_tasks: Sequence[DatasetProducingTask]

    def is_produced_by_task(self, *, task_id: str, dag_id: str) -> bool:
        return any(
            task.task_id == task_id and task.dag_id == dag_id for task in self.producing_tasks
        )


###################################################################################################
# Serialized data that scopes to airflow DAGs and tasks.
###################################################################################################
# History:
# - created
@whitelist_for_serdes
@record
class SerializedDagData:
    """A record containing pre-computed data about a given airflow dag."""

    dag_id: str
    dag_info: DagInfo
    source_code: Optional[str]
    leaf_asset_keys: set[AssetKey]
    task_infos: Mapping[str, TaskInfo]


@whitelist_for_serdes
@record
class KeyScopedTaskHandles:
    asset_key: AssetKey
    mapped_tasks: AbstractSet[TaskHandle]


@whitelist_for_serdes
@record
class KeyScopedDagHandles:
    asset_key: AssetKey
    mapped_dags: AbstractSet[DagHandle]


###################################################################################################
# Serializable data that will be cached to avoid repeated calls to the Airflow API, and to avoid
# repeated scans of passed-in Definitions objects.
###################################################################################################
# History:
# - created
# - removed existing_asset_data
# - added key_scope_data_items
# - added instance_name
@whitelist_for_serdes
@record
class SerializedAirflowDefinitionsData:
    instance_name: str
    key_scoped_task_handles: list[KeyScopedTaskHandles]
    key_scoped_dag_handles: list[KeyScopedDagHandles]
    dag_datas: Mapping[str, SerializedDagData]
    datasets: Sequence[Dataset]

    @cached_property
    def all_mapped_tasks(self) -> dict[AssetKey, AbstractSet[TaskHandle]]:
        return {item.asset_key: item.mapped_tasks for item in self.key_scoped_task_handles}

    @cached_property
    def all_mapped_dags(self) -> dict[AssetKey, AbstractSet[DagHandle]]:
        return {item.asset_key: item.mapped_dags for item in self.key_scoped_dag_handles}

    @cached_property
    def datasets_by_dag_id(self) -> dict[str, set[str]]:
        """Mapping of dag_id to set of dataset uris it produces, in any task."""
        dataset_map: dict[str, set[str]] = defaultdict(set)
        for dataset in self.datasets:
            for producing_task in dataset.producing_tasks:
                dataset_map[producing_task.dag_id].add(dataset.uri)
        return dataset_map

    @cached_property
    def upstream_datasets_by_uri(self) -> dict[str, set[str]]:
        """Mapping of dataset uri to set of upstream dataset uris."""
        upstream_datasets_map: dict[str, set[str]] = defaultdict(set)
        for dataset in self.datasets:
            for consuming_dag_id in dataset.consuming_dags:
                datasets_produced_by_consuming_dag = self.datasets_by_dag_id.get(
                    consuming_dag_id.dag_id, set()
                )
                for downstream_dataset_uri in datasets_produced_by_consuming_dag:
                    upstream_datasets_map[downstream_dataset_uri].add(dataset.uri)
        return upstream_datasets_map



from collections import defaultdict
from collections.abc import Mapping, Sequence
from functools import cached_property
from typing import AbstractSet, Union  # noqa: UP035

from dagster import AssetKey, AssetsDefinition, AssetSpec
from dagster._annotations import beta, public
from dagster._core.definitions.job_definition import JobDefinition
from dagster._core.definitions.repository_definition.repository_definition import (
    RepositoryDefinition,
)
from dagster._record import record

from dagster_airlift.core.airflow_instance import AirflowInstance
from dagster_airlift.core.serialization.compute import (
    AirliftMetadataMappingInfo,
    build_airlift_metadata_mapping_info,
)
from dagster_airlift.core.serialization.serialized_data import DagHandle, TaskHandle
from dagster_airlift.core.utils import (
    dag_handle_from_job,
    dag_handles_for_spec,
    get_producing_dag_ids,
    is_airflow_mapped_job,
    is_dag_mapped_asset_spec,
    is_peered_dag_asset_spec,
    is_task_mapped_asset_spec,
    peered_dag_handles_for_spec,
    spec_iterator,
    task_handles_for_spec,
)

MappedAsset = Union[AssetSpec, AssetsDefinition]


def _is_mapped_asset_spec(asset: AssetSpec) -> bool:
    return (
        is_task_mapped_asset_spec(asset)
        or is_dag_mapped_asset_spec(asset)
        or is_peered_dag_asset_spec(asset)
    )


@beta
@record
class AirflowDefinitionsData:
    """A class that holds data about the assets that are mapped to Airflow dags and tasks, and
    provides methods for retrieving information about the mappings.
    The user should not instantiate this class directly. It is provided when customizing the events
    that are generated by the Airflow sensor using the `event_transformer_fn` argument of
    :py:func:`build_defs_from_airflow_instance`.
    """

    airflow_instance: AirflowInstance
    resolved_repository: RepositoryDefinition

    @property
    def airflow_mapped_asset_specs(self) -> Mapping[AssetKey, AssetSpec]:
        """The assets that are mapped to Airflow tasks and dags."""
        return {
            spec.key: spec
            for spec in spec_iterator(self.resolved_repository.assets_defs_by_key.values())
            if _is_mapped_asset_spec(spec)
        }

    @property
    def airflow_mapped_jobs(self) -> Sequence[JobDefinition]:
        """Jobs mapping to Airflow dags."""
        return [
            job for job in self.resolved_repository.get_all_jobs() if is_airflow_mapped_job(job)
        ]

    @property
    def airflow_mapped_jobs_by_dag_handle(
        self,
    ) -> Mapping[DagHandle, JobDefinition]:
        """Jobs mapping to Airflow dags by dag_id."""
        return {dag_handle_from_job(job): job for job in self.airflow_mapped_jobs}

    @property
    def assets_per_job(self) -> Mapping[str, AbstractSet[AssetKey]]:
        """Assets per job mapping to Airflow dags."""
        return {
            job.name: self.assets_produced_by_dags[dag_handle.dag_id]
            for dag_handle, job in self.airflow_mapped_jobs_by_dag_handle.items()
        }

    @property
    def assets_produced_by_dags(self) -> Mapping[str, AbstractSet[AssetKey]]:
        """Assets produced by Airflow dags."""
        result = defaultdict(set)
        for spec in self.airflow_mapped_asset_specs.values():
            for dag_id in get_producing_dag_ids(spec):
                result[dag_id].add(spec.key)
        return result

    @public
    @property
    def instance_name(self) -> str:
        """The name of the Airflow instance."""
        return self.airflow_instance.name

    @cached_property
    def mapping_info(self) -> AirliftMetadataMappingInfo:
        return build_airlift_metadata_mapping_info(self.airflow_mapped_asset_specs.values())

    @public
    def task_ids_in_dag(self, dag_id: str) -> set[str]:
        """Returns the task ids within the given dag_id.

        Args:
            dag_id (str): The dag id.
        """
        return self.mapping_info.task_id_map[dag_id]

    @property
    def dag_ids_with_mapped_asset_keys(self) -> AbstractSet[str]:
        """All dag_ids that have asset keys explicitly mapped to them. This include peered dag assets."""
        # dag ids that have asset keys explicitly mapped to them, or tasks within them
        explicitly_mapped_dag_ids = self.mapping_info.dag_ids
        # dag ids that have a "peered" dag asset
        peered_dag_ids = {
            handle.dag_id for handle in self.peered_dag_asset_keys_by_dag_handle.keys()
        }
        return explicitly_mapped_dag_ids.union(peered_dag_ids)

    @cached_property
    def mapped_asset_keys_by_task_handle(self) -> Mapping[TaskHandle, AbstractSet[AssetKey]]:
        asset_keys_per_handle = defaultdict(set)
        for spec in self.airflow_mapped_asset_specs.values():
            if is_task_mapped_asset_spec(spec):
                task_handles = task_handles_for_spec(spec)
                for task_handle in task_handles:
                    asset_keys_per_handle[task_handle].add(spec.key)
        return asset_keys_per_handle

    # these dag handle properties are ripe for consolidation
    @cached_property
    def mapped_asset_keys_by_dag_handle(self) -> Mapping[DagHandle, AbstractSet[AssetKey]]:
        """Assets specifically mapped to each dag."""
        asset_keys_per_handle = defaultdict(set)
        for spec in self.airflow_mapped_asset_specs.values():
            if is_dag_mapped_asset_spec(spec):
                dag_handles = dag_handles_for_spec(spec)
                for dag_handle in dag_handles:
                    asset_keys_per_handle[dag_handle].add(spec.key)
        return asset_keys_per_handle

    @cached_property
    def peered_dag_asset_keys_by_dag_handle(self) -> Mapping[DagHandle, AbstractSet[AssetKey]]:
        """Autogenerated "peered" dag assets."""
        asset_keys_per_handle = defaultdict(set)
        for spec in self.airflow_mapped_asset_specs.values():
            if is_peered_dag_asset_spec(spec):
                dag_handles = peered_dag_handles_for_spec(spec)
                for dag_handle in dag_handles:
                    asset_keys_per_handle[dag_handle].add(spec.key)
        return asset_keys_per_handle

    @cached_property
    def all_asset_keys_by_dag_handle(self) -> Mapping[DagHandle, AbstractSet[AssetKey]]:
        """All asset keys mapped to each dag."""
        res = defaultdict(set)
        for handle, keys in self.mapped_asset_keys_by_dag_handle.items():
            res[handle].update(keys)
        for handle, keys in self.peered_dag_asset_keys_by_dag_handle.items():
            res[handle].update(keys)
        return res

    @public
    def asset_keys_in_task(self, dag_id: str, task_id: str) -> AbstractSet[AssetKey]:
        """Returns the asset keys that are mapped to the given task.

        Args:
            dag_id (str): The dag id.
            task_id (str): The task id.
        """
        return self.mapped_asset_keys_by_task_handle[TaskHandle(dag_id=dag_id, task_id=task_id)]



import inspect
import json
import os
from collections.abc import Callable, Iterable, Mapping, Sequence
from typing import Any

import requests
from airflow.models import BaseOperator

from dagster_airlift.constants import DEFER_ASSET_EVENTS_TAG, TASK_MAPPING_METADATA_KEY
from dagster_airlift.in_airflow.base_asset_operator import BaseDagsterAssetsOperator, Context


class BaseProxyTaskToDagsterOperator(BaseDagsterAssetsOperator):
    """An operator that proxies task execution to Dagster assets with metadata that map to this task's dag ID and task ID.

    For the DAG ID and task ID that this operator proxies, it expects there to be corresponding assets
    in the linked Dagster deployment that have metadata entries with the key `dagster-airlift/task-mapping` that
    map to this DAG ID and task ID. This metadata is typically set using the
    :py:func:`dagster_airlift.core.assets_with_task_mappings` function.

    The following methods must be implemented by subclasses:

        - :py:meth:`get_dagster_session` (inherited from :py:class:`BaseDagsterAssetsOperator`)
        - :py:meth:`get_dagster_url` (inherited from :py:class:`BaseDagsterAssetsOperator`)
        - :py:meth:`build_from_task` A class method which takes the task to be proxied, and constructs
            an instance of this operator from it.

    There is a default implementation of this operator, :py:class:`DefaultProxyTaskToDagsterOperator`,
    which is used by :py:func:`proxying_to_dagster` if no override operator is provided.
    """

    @property
    def should_defer_asset_events(self) -> bool:
        return False

    def filter_asset_nodes(
        self, context: Context, asset_nodes: Sequence[Mapping[str, Any]]
    ) -> Iterable[Mapping[str, Any]]:
        for asset_node in asset_nodes:
            if matched_dag_id_task_id(
                asset_node, self.get_airflow_dag_id(context), self.get_airflow_task_id(context)
            ):
                yield asset_node

    @classmethod
    def build_from_task(cls, task: BaseOperator) -> "BaseProxyTaskToDagsterOperator":
        return build_dagster_task(task, cls)

    def default_dagster_run_tags(self, context: Context) -> dict[str, str]:
        tags = super().default_dagster_run_tags(context)
        if self.should_defer_asset_events:
            tags[DEFER_ASSET_EVENTS_TAG] = "true"
        return tags


class DefaultProxyTaskToDagsterOperator(BaseProxyTaskToDagsterOperator):
    """The default task proxying operator - which opens a blank session and expects the dagster URL to be set in the environment.
    The dagster url is expected to be set in the environment as DAGSTER_URL.

    This operator should not be instantiated directly - it is instantiated by :py:func:`proxying_to_dagster` if no
    override operator is provided.
    """

    def get_dagster_session(self, context: Context) -> requests.Session:
        return requests.Session()

    def get_dagster_url(self, context: Context) -> str:
        return os.environ["DAGSTER_URL"]


def build_dagster_task(
    original_task: BaseOperator,
    dagster_operator_klass: type[BaseProxyTaskToDagsterOperator],
) -> BaseProxyTaskToDagsterOperator:
    return instantiate_dagster_operator(original_task, dagster_operator_klass)


def instantiate_dagster_operator(
    original_task: BaseOperator,
    dagster_operator_klass: type[BaseProxyTaskToDagsterOperator],
) -> BaseProxyTaskToDagsterOperator:
    """Instantiates a DagsterOperator as a copy of the provided airflow task.

    We attempt to copy as many of the original task's attributes as possible, while respecting
    that attributes may change between airflow versions. In order to do this, we inspect the
    arguments available to the BaseOperator constructor and copy over any of those arguments that
    are available as attributes on the original task.
    This approach has limitations:
    - If the task attribute is transformed and stored on another property, it will not be copied.
    - If the task attribute is transformed in a way that makes it incompatible with the constructor arg
    and stored in the same property, that will attempt to be copied and potentiall break initialization.
    In the future, if we hit problems with this, we may need to add argument overrides to ensure we either
    attempt to include certain additional attributes, or exclude others. If this continues to be a problem
    across airflow versions, it may be necessary to revise this approach to one that explicitly maps airflow
    version to a set of expected arguments and attributes.
    """
    base_operator_args, base_operator_args_with_defaults = get_params(BaseOperator.__init__)
    init_kwargs = {}

    ignore_args = [
        # These don't make sense in context to copy
        "kwargs",
        "args",
        "dag",
        # The weight rule stored on the base operator is a private subclass of PriorityWeightStrategy,
        # which satisfies the type signature of the constructor, but fails the validation process in
        # the constructor. See https://github.com/apache/airflow/blob/2b15e9f26fee27b6c1fbc8167d0e0558198ffa7a/airflow/task/priority_strategy.py#L127
        # for more details.
        # We could likely add custom handling here to support the parameter.
        # For now, we ignore it, as it's currently an experimental feature in Airflow.
        "weight_rule",
    ]
    for arg in base_operator_args:
        if arg in ignore_args or getattr(original_task, arg, None) is None:
            continue
        init_kwargs[arg] = getattr(original_task, arg)
    for kwarg, default in base_operator_args_with_defaults.items():
        if kwarg in ignore_args or getattr(original_task, kwarg, None) is None:
            continue
        init_kwargs[kwarg] = getattr(original_task, kwarg, default)

    # Make sure that the operator overrides take precedence.

    return dagster_operator_klass(**init_kwargs)


def get_params(func: Callable[..., Any]) -> tuple[set[str], dict[str, Any]]:
    """Retrieves the args and kwargs from the signature of a given function or method.
    For kwargs, default values are retrieved as well.

    Args:
        func (Callable[..., Any]): The function or method to inspect.

    Returns:
        Tuple[Set[str], Dict[str, Any]]:
            - A set of argument names that do not have default values.
            - A dictionary of keyword argument names and their default values.
    """
    # Get the function's signature
    sig = inspect.signature(func)

    # Initialize sets for args without defaults and kwargs with defaults
    args_with_defaults = {}
    args = set()

    # Iterate over function parameters
    for name, param in sig.parameters.items():
        if param.default is inspect.Parameter.empty and name != "self":  # Exclude 'self'
            args.add(name)
        else:
            if name != "self":  # Exclude 'self'
                args_with_defaults[name] = param.default

    return args, args_with_defaults


def matched_dag_id_task_id(asset_node: Mapping[str, Any], dag_id: str, task_id: str) -> bool:
    json_metadata_entries = {
        entry["label"]: entry["jsonString"]
        for entry in asset_node["metadataEntries"]
        if entry["__typename"] == "JsonMetadataEntry"
    }

    mapping_entry = json_metadata_entries.get(TASK_MAPPING_METADATA_KEY)
    if mapping_entry:
        task_handle_dict_list = json.loads(mapping_entry)
        for task_handle_dict in task_handle_dict_list:
            if task_handle_dict["dag_id"] == dag_id and task_handle_dict["task_id"] == task_id:
                return True

    return False



import datetime
from typing import Optional, Union

from dagster import RunRequest, sensor
from dagster._annotations import beta
from dagster._config.pythonic_config.config import Config
from dagster._core.definitions.decorators.job_decorator import job
from dagster._core.definitions.decorators.op_decorator import op
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.definitions.job_definition import JobDefinition
from dagster._core.definitions.op_definition import OpDefinition
from dagster._core.definitions.run_config import RunConfig
from dagster._core.definitions.run_request import SkipReason
from dagster._core.definitions.sensor_definition import (
    DefaultSensorStatus,
    SensorDefinition,
    SensorEvaluationContext,
)
from dagster._core.execution.context.op_execution_context import OpExecutionContext
from dagster._core.storage.dagster_run import DagsterRun, RunsFilter
from dagster._daemon.auto_run_reexecution.auto_run_reexecution import run_was_successfully_retried
from dagster._grpc.client import DEFAULT_SENSOR_GRPC_TIMEOUT
from dagster._record import record
from dagster._serdes import deserialize_value, serialize_value
from dagster._time import get_current_datetime
from dagster_shared.serdes import whitelist_for_serdes
from pydantic import Field

from dagster_airlift.core.airflow_defs_data import AirflowDefinitionsData
from dagster_airlift.core.airflow_instance import AirflowInstance
from dagster_airlift.core.monitoring_job.event_stream import persist_events
from dagster_airlift.core.monitoring_job.utils import structured_log
from dagster_airlift.core.utils import monitoring_job_name

MAIN_LOOP_TIMEOUT_SECONDS = DEFAULT_SENSOR_GRPC_TIMEOUT - 20
DEFAULT_AIRFLOW_SENSOR_INTERVAL_SECONDS = 30
NO_CURSOR_LOOKBACK_DELTA = datetime.timedelta(days=1)


@whitelist_for_serdes
@record
class AirflowMonitoringJobSensorCursor:
    range_start: str
    range_end: str

    def to_config(self) -> "MonitoringConfig":
        return MonitoringConfig(range_start=self.range_start, range_end=self.range_end)

    def to_tags(self) -> dict[str, str]:
        return {
            "range_start": self.range_start,
            "range_end": self.range_end,
        }

    def advance(self, effective_timestamp: datetime.datetime) -> "AirflowMonitoringJobSensorCursor":
        return AirflowMonitoringJobSensorCursor(
            range_start=self.range_end, range_end=effective_timestamp.isoformat()
        )


# IMPROVEME BCOR-102: We should be able to replace the sensor from the original Airlift functionality with this job.
@beta
def build_airflow_monitoring_defs(
    *,
    airflow_instance: AirflowInstance,
) -> Definitions:
    """The constructed job polls the Airflow instance for activity, and inserts asset events into Dagster's event log."""
    return Definitions(
        jobs=[build_monitoring_job(airflow_instance=airflow_instance)],
        sensors=[build_monitoring_sensor(airflow_instance=airflow_instance)],
    )


class MonitoringConfig(Config):
    range_start: str = Field(
        description="The start of the range to process. Should be an ISO 8601 formatted string."
    )
    range_end: str = Field(
        description="The end of the range to process. Should be an ISO 8601 formatted string."
    )

    @property
    def range_start_datetime(self) -> datetime.datetime:
        return datetime.datetime.fromisoformat(self.range_start)

    @property
    def range_end_datetime(self) -> datetime.datetime:
        return datetime.datetime.fromisoformat(self.range_end)


def build_monitoring_sensor(
    *,
    airflow_instance: AirflowInstance,
) -> SensorDefinition:
    @sensor(
        job_name=monitoring_job_name(airflow_instance.name),
        name=f"{airflow_instance.name}__airflow_monitoring_job_sensor",
        default_status=DefaultSensorStatus.RUNNING,
    )
    def airflow_monitoring_job_sensor(
        context: SensorEvaluationContext,
    ) -> Union[RunRequest, SkipReason]:
        effective_timestamp = get_current_datetime()
        if context.cursor is None:
            cursor = AirflowMonitoringJobSensorCursor(
                range_start=(get_current_datetime() - NO_CURSOR_LOOKBACK_DELTA).isoformat(),
                range_end=effective_timestamp.isoformat(),
            )
        else:
            cursor = deserialize_value(context.cursor, AirflowMonitoringJobSensorCursor)

        run = _get_run_for_cursor(context, airflow_instance, cursor)
        # We only advance the cursor if the run has finished successfully.
        if run and not run.is_finished:
            return SkipReason(
                f"Monitoring job is still running for range {cursor.range_start} to {cursor.range_end}. Waiting to advance."
            )
        if run and not run.is_success and not run_was_successfully_retried(run, context.instance):
            raise Exception(
                f"Monitoring job failed for range {cursor.range_start} to {cursor.range_end} with run {run.run_id}. Dagster currently expects runs to be inserted in time-increasing order. To avoid unexpected side effects, the sensor will wait until this range has been successfully executed before advancing. To rectify this, you can either wait for automatic retries to succeed, or you can manually re-execute the failed run once the issue is resolved."
            )
        cursor = cursor if not run else cursor.advance(effective_timestamp)
        context.update_cursor(serialize_value(cursor))

        return RunRequest(
            run_config=RunConfig(
                ops={monitoring_job_op_name(airflow_instance): cursor.to_config()},
            ),
            tags=cursor.to_tags(),
        )

    return airflow_monitoring_job_sensor


def _get_run_for_cursor(
    context: SensorEvaluationContext,
    airflow_instance: AirflowInstance,
    cursor: AirflowMonitoringJobSensorCursor,
) -> Optional[DagsterRun]:
    return next(
        iter(
            context.instance.get_runs(
                filters=RunsFilter(
                    job_name=monitoring_job_name(airflow_instance.name), tags=cursor.to_tags()
                ),
                limit=1,
            )
        ),
        None,
    )


def _build_monitoring_op(
    airflow_instance: AirflowInstance,
) -> OpDefinition:
    @op(
        name=monitoring_job_op_name(airflow_instance),
    )
    def monitor_dags(context: OpExecutionContext, config: MonitoringConfig) -> None:
        """The main function that runs the sensor. It polls the Airflow instance for activity and emits asset events."""
        # This is a hack to get the repository tag for the current run. It's bad because it assumes that the job we're
        # creating a run for is within the same repository; but I think that we'll have to do a second pass to get "outside of code
        # location" runs working (if that's even something we want to do).
        airflow_data = AirflowDefinitionsData(
            airflow_instance=airflow_instance, resolved_repository=context.repository_def
        )

        structured_log(
            context,
            f"Processing from {config.range_start} to {config.range_end}",
        )
        persist_events(
            context,
            airflow_data,
            airflow_instance,
            config.range_start_datetime.timestamp(),
            config.range_end_datetime.timestamp(),
        )

    return monitor_dags


def monitoring_job_op_name(airflow_instance: AirflowInstance) -> str:
    return f"core_monitor__{airflow_instance.name}"


def build_monitoring_job(
    *,
    airflow_instance: AirflowInstance,
) -> JobDefinition:
    @job(name=monitoring_job_name(airflow_instance.name))
    def airflow_monitoring_job():
        _build_monitoring_op(airflow_instance)()

    return airflow_monitoring_job



from collections.abc import Mapping
from typing import Any, NamedTuple, Optional

# These constants exist within the Dagster main package, but are duplicated here to avoid taking a dependency on Dagster within
# the dagster-airlift[in-airflow] submodule.
TERMINAL_STATI = ["SUCCESS", "FAILURE", "CANCELED"]
SYSTEM_TAG_PREFIX = "dagster/"
MAX_RETRIES_TAG = f"{SYSTEM_TAG_PREFIX}max_retries"
WILL_RETRY_TAG = f"{SYSTEM_TAG_PREFIX}will_retry"
AUTO_RETRY_RUN_ID_TAG = f"{SYSTEM_TAG_PREFIX}auto_retry_run_id"
RETRY_NUMBER_TAG = f"{SYSTEM_TAG_PREFIX}retry_number"
PARENT_RUN_ID_TAG = f"{SYSTEM_TAG_PREFIX}parent_run_id"
SUCCESS_STATUS = "SUCCESS"
RETRY_ON_ASSET_OR_OP_FAILURE_TAG = f"{SYSTEM_TAG_PREFIX}retry_on_asset_or_op_failure"
RUN_FAILURE_REASON_TAG = f"{SYSTEM_TAG_PREFIX}failure_reason"
STEP_FAILURE_REASON = "STEP_FAILURE"


class DagsterRunResult(NamedTuple):
    status: str
    tags: Mapping[str, Any]

    @property
    def run_will_automatically_retry(self) -> bool:
        return (
            not self.succeeded
            and get_boolean_tag_value(self.tags.get(WILL_RETRY_TAG), False) is True
        )

    @property
    def retried_run_id(self) -> Optional[str]:
        return self.tags.get(AUTO_RETRY_RUN_ID_TAG)

    @property
    def succeeded(self) -> bool:
        return self.status == SUCCESS_STATUS


def get_boolean_tag_value(tag_value: Optional[str], default_value: bool = False) -> bool:
    if tag_value is None:
        return default_value

    return tag_value.lower() not in {"false", "none", "0", ""}
